/* x16 Raycaster demo in c - by Jeffrey - 2021 */

#include <stdint.h>
#include <stdio.h>  // e.g. printf()
#include <conio.h>  // e.g. clrscr()
#include <cbm.h>    // e.g. cbm_k_bsout()
#include <joystick.h>
#include <cx16.h>   // e.g. VERA.control

#define SCREEN_WIDTH 320
#define SCREEN_HEIGHT (200+40)   // The original Wolfenstein was 320x200
#define PAL_SCREEN_HEIGHT 240
#define RENDER_HEIGHT (152+30)
#define RENDER_WIDTH 304

#define TEXTURES_VRAM_BASE 0x8C00 // NOTE: its 0x18C00 (with high bit set) -> this is right behind the BITMAP memory!

#define WALL_HEIGHT_NUMERATOR 8500
#define CEILING_COLOR 19
#define FLOOR_COLOR 22

#define PROGRESS_BAR_HEIGHT 10
#define PROGRESS_BAR_BORDER_HEIGHT 10
#define PROGRESS_BAR_BORDER_WIDTH 10

#define TOP_BORDER_HEIGHT 4
#define MIDDLE_BORDER_HEIGHT 8
#define BOTTOM_BORDER_HEIGHT 3
#define BORDER_WIDTH ((SCREEN_WIDTH-RENDER_WIDTH)/2)
#define NR_OF_ANGLES (6*RENDER_WIDTH)  // This is the number of angles we use in a full circle (so 304 angles per 60 degrees)
#define TILE_SIZE 256            // width and height in pixels
#define SEARCH_STEP_FRACTION 16  // the fraction of the tile size we increase our ray each search-step (increasing makes the borders more smooth, costing performance)

#include "ray.h"

/*
 1 = Blue Stone 1
 2 = Blue Stone 2
 3 = BS cell
 4 = BS cell skeleton
 5 = Door side OR Blue Stone 1
 6 = Closed door

*/

#define MAP_HEIGHT 14
#define MAP_WIDTH 15
u8 world_map[MAP_HEIGHT][MAP_WIDTH]=
{
  {1,1,2,1,2,1,1,6,1,1,1,1,1,2,1},
  {2,0,0,0,0,5,0,0,0,5,0,0,0,0,2},
  {1,0,0,0,0,0,0,0,0,0,0,0,0,0,2},
  {1,0,0,0,0,5,0,0,0,5,0,0,0,0,2},
  {2,0,0,0,0,2,0,0,0,2,0,0,0,0,2},
  {1,1,2,1,1,2,0,0,0,1,1,1,1,1,1},
  {1,0,0,0,0,5,0,0,0,1,0,0,0,0,1},
  {2,0,0,0,0,0,0,0,0,6,0,0,0,0,2},
  {1,0,0,0,0,5,0,0,0,1,0,0,0,0,1},
  {1,2,2,1,1,2,0,0,0,2,1,1,2,1,1},
  {1,0,0,0,0,0,0,0,0,0,0,0,0,0,1},
  {2,0,0,0,0,0,0,0,0,0,0,0,0,0,1},
  {1,0,0,0,0,0,0,0,0,0,0,0,0,0,1},
  {1,3,2,1,3,1,2,4,1,2,3,2,1,3,1},
};


/*
#define MAP_HEIGHT 9
#define MAP_WIDTH 14
u8 world_map[MAP_HEIGHT][MAP_WIDTH]=
{
  {2,2,2,3,3,3,3,4,4,4,4,4,4,4},
  {2,0,0,0,0,0,4,0,0,0,0,0,0,5},
  {2,0,0,0,0,0,4,0,0,0,0,0,0,5},
  {2,2,2,0,4,4,4,4,6,0,0,0,0,5},
  {3,0,0,0,0,0,0,0,6,0,2,2,0,5},
  {3,0,0,0,0,0,4,0,6,0,2,2,0,5},
  {3,0,0,3,0,0,4,0,6,0,0,0,0,5},
  {3,0,0,3,0,0,4,0,0,0,0,0,0,5},
  {3,3,3,3,3,3,3,7,7,7,7,7,7,7},
};
*/
/*
#define MAP_HEIGHT 5
#define MAP_WIDTH 5
u8 world_map[MAP_HEIGHT][MAP_WIDTH]=
{
  {1,1,3,3,1},
  {1,0,0,0,2},
  {1,0,0,0,5},
  {1,0,0,0,6},
  {1,1,4,4,1},
};
*/

u8 * world_map_c; 

//u8 use_dda = 1; // Digital differential analyzer (= what Wolfenstein 3D used)
u8 use_asm_draw_pure = 1;
u8 use_asm_draw = 1; // Draw vertical bar in assembly
u8 use_textures = 1;

u8 tile_value = 0; // FIXME: we use a global here because we can't return multiple values. This is just a quick and dirty workaround.
u8 hit_point_on_wall = 0; // FIXME: we use a global here because we can't return multiple values. This is just a quick and dirty workaround.

u16 byte_code_index = 0;
u16 palette_index = 0;
    
void main(void) {
    
    u16 /*distance_to_wall, */ normalized_distance_to_wall;
    u8 background_color, stats_background_color, bar_color;
    u16 text_y, text_x;
    u16 bar_x_start, wall_height, capped_wall_height;
    
    i16 player_pos_x, player_pos_y, sin_player_dir, cos_player_dir;
    i16 player_dir_index, ray_dir_index, test_ray_dir_index;
    i16 ray_dir_index_offset;
    // i16 ray_vs_player_angle_diff;
    i16 x_move, y_move;
    
    u8 show_test_ray;
    u8 light_offset;
    u8 key_pressed;
    u8 do_print = 0;
    
    u8 joystick_install_result = 0;
    u8 joystick_count = 0;
    u8 joystick0_state = 0;
    
    u8 debug_is_enabled = 0;
    u8 start_ticks;
    
    // TODO: if joystick_install_result is not ok, we should do something with that!
    joystick_install_result = joy_install(&joy_static_stddrv);
    
    // Active vsync irq handler (written in ray.asm)
    _enable_vsync_counter();
    
    copy_world_map_to_four_quadrants();
    
    clrscr();
    cbm_k_bsout(CH_FONT_UPPER);
    move_text_screen_in_memory();
    
    init_draw_screen();
    set_palette_color(3, 0, 4, 4);
    background_color = 3;
    stats_background_color = 197;
    clear_screen();
    
    _use_textures = use_textures;
    draw_background(background_color, stats_background_color);
    
    text_x = 0*8;
    text_y = 3*8;
    draw_text("                controls  ", text_x, text_y, 11);
    text_y += 10;
    text_y += 10;
    text_y += 10;
    draw_text("              up - move forwards", text_x, text_y, 1);
    text_y += 10;
    draw_text("            down - move backwards", text_x, text_y, 1);
    text_y += 10;
    text_y += 10;
    draw_text("            left - turn left", text_x, text_y, 1);
    text_y += 10;
    draw_text("           right - turn right", text_x, text_y, 1);
    text_y += 10;
    text_y += 10;
    draw_text("        alt+left - strafe left", text_x, text_y, 1);
    text_y += 10;
    draw_text("       alt+right - strafe right", text_x, text_y, 1);
    text_y += 10;
    text_y += 10;
    text_y += 10;
    draw_text("            - press any key -", text_x, text_y, 2);
    
    draw_text("raycaster demo v1.3.0", 10*8, 25*8 + 4, 1);
    draw_text("by jeffrey", 16*8, 27*8 + 4, 1);

    enable_display_layer(0);
    disable_display_layer(1);
    make_layer_1_transparent();
    
    while (!kbhit());
    cgetc(); // ignoring the pressed key
    clrscr();
    clear_render_area();
    
    
    // Initial position (and direction) of the player
    /*
    // == 5 x 5 map ==
    player_pos_x = 256 * 2 + 128;
    player_pos_y = 256 * 1 + 128;
    //player_dir_index = (NR_OF_ANGLES*1)/8;
    player_dir_index = (NR_OF_ANGLES*3)/12 + 12; // 360 / 12 = 30 degrees
    */
    
    // == Other test map ==
    //player_pos_x = 256 * 3 + 128;
    //player_pos_y = 256 * 7 + 20;
    //player_pos_x = 256 * 2 + 53;
    //player_pos_y = 256 * 2 + 73;
    // player_dir_index = (NR_OF_ANGLES*1)/8 + 23;
    
    // == Episode 1, level 1 map ==
    
    // Official starting position
    player_pos_x = 256 * 2 + 128;
    player_pos_y = 256 * 6 + 128;
    player_dir_index = (NR_OF_ANGLES*0)/4;

    /*
    // Nice location
    player_pos_x = 256 * 13 + 128;
    player_pos_y = 256 * 3 + 128;
    player_dir_index = (NR_OF_ANGLES*2)/4;
    */

    test_ray_dir_index = player_dir_index;
    show_test_ray = 0;
    
    if (use_asm_draw) {
        // We still need _init_draw_code because it is used to draw the ceiling and floor!
        _init_draw_code();
        if (use_textures) {
            // NOTE: we start at palette index 8*16. Each texture needs a certain amount of colors. Every time we call copy_texture_to_vram palette_index is increased.
            palette_index = 2*16;
            // Filenames here need to be lowecase here, on the filesystem it needs to be UPPERCASE
            load_file_into_banked_ram("bluestone1.bin"); 
            copy_texture_to_vram(0);
            load_file_into_banked_ram("bluestone2.bin");  // TODO: BS1 and BS2 can probably share their palettes!!
            copy_texture_to_vram(1);
            load_file_into_banked_ram("bscell.bin");
            copy_texture_to_vram(2);
            load_file_into_banked_ram("bscellskeleton.bin");
            copy_texture_to_vram(3);
            load_file_into_banked_ram("doorside.bin");
            copy_texture_to_vram(4);
            load_file_into_banked_ram("closeddoor.bin");
            copy_texture_to_vram(5);
            _generate_texture_draw_code();
        }
    }
    else {
        if (use_textures) {
            // NOTE: we only read in one texture when not in asm mode: we only have 1 banked ram to use
            // NOTE: we start at palette index 8*16. Each texture needs a certain amount of colors. Every time we call copy_texture_to_vram palette_index is increased.
            palette_index = 8*16;
            load_file_into_banked_ram("bluestone1.bin"); // This needs to be lowecase here, on the filesystem it needs to be UPPERCASE
            // NOTE: we call this function because it also sets the palette (even though we do not use the texture in VRAM)
            copy_texture_to_vram(0);
        }
    }
    
    if (!use_textures) {
        create_new_palette();
    }


    while(1) {
        start_ticks = _ticks;

        if (use_asm_draw_pure) {
            _player_pos_x = player_pos_x;
            _player_pos_y = player_pos_y;
            // FIXME: we should do this in assembly (and only when player_dir_index is changing)
            _sin_player_dir = sine(player_dir_index);
            _cos_player_dir = cosine(player_dir_index);
            
            _start_ray_dir_index = player_dir_index - NR_OF_ANGLES/12;  // minus 30 degrees
            // this should be positive, so we make it so
            if (_start_ray_dir_index < 0) {
                _start_ray_dir_index = _start_ray_dir_index + NR_OF_ANGLES;
            }
            _raycast_and_draw();
        }
        else if (use_asm_draw) {
            

            sin_player_dir = sine(player_dir_index);
            cos_player_dir = cosine(player_dir_index);
            
            for (ray_dir_index_offset = -NR_OF_ANGLES/12; ray_dir_index_offset < NR_OF_ANGLES/12; ray_dir_index_offset++) {
            // for (ray_dir_index_offset = 20; ray_dir_index_offset < 21; ray_dir_index_offset++) {

                if (player_dir_index + ray_dir_index_offset < 0) {
                    // Important: the modulo in c is a *remained*-function. So it doesn't work like modulo for *negative* numbers! Therefore we have to add NR_OF_ANGLES first.
                    ray_dir_index = (player_dir_index + ray_dir_index_offset + NR_OF_ANGLES) % NR_OF_ANGLES;
                }
                else {
                    ray_dir_index = (player_dir_index + ray_dir_index_offset) % NR_OF_ANGLES;
                }
                               
                normalized_distance_to_wall = get_distance_to_wall_dda_asm(player_pos_x, player_pos_y, ray_dir_index, sin_player_dir, cos_player_dir);
                // normalized_distance_to_wall = get_distance_to_wall_dda(player_pos_x, player_pos_y, ray_dir_index, sin_player_dir, cos_player_dir, do_print);
                //enable_display_layer(1);
                //printf("cycles: %d\n", _cycle_end - _cycle_start);
                //while(1);
                
                wall_height = WALL_HEIGHT_NUMERATOR / (normalized_distance_to_wall / 8); // This is square!
                
                if (!use_textures) {
                    // Using sine now, kinda works nicely
                    // TODO: make this better: 
                    light_offset = 3 + sine(wall_height*(NR_OF_ANGLES/4)/1024) / 8;
                    //light_offset = 1 + wall_height / 16;
                    if (light_offset > 15) {
                        light_offset = 15;
                    }
                    // TODO: we are using tile_value (=global) here. This is a quick-and-dirty global to get the value of a tile that was just hit
                    bar_color = tile_value * 16 + light_offset;
                    if (show_test_ray && test_ray_dir_index == ray_dir_index) {
                        bar_color = 1;
                    }
                    _color_to_draw = bar_color;
                }
                else {
                    _color_to_draw = tile_value;
                }
                
                bar_x_start = RENDER_WIDTH/2 - 1 - ray_dir_index_offset;
                
                // TODO: Its probably better to draw from left to right, so we should instead iterate from high to low indexes
                if (wall_height > RENDER_HEIGHT) {
                    capped_wall_height = RENDER_HEIGHT;
                }
                else {
                    capped_wall_height = wall_height;
                }

                _pixel_column = bar_x_start;
                _capped_wall_height  = capped_wall_height;

                // TODO: we are now capping the wall_height. We don't want to do that. But we are limited to 512 different heigts atm.
                if (wall_height > 511) {
                    wall_height = 511;
                }
                _wall_height = wall_height;
                _draw_vertical_bar();

            }
            
        }
        /*
        else {
            sin_player_dir = sine(player_dir_index);
            cos_player_dir = cosine(player_dir_index);
            
            // TODO: for now we make sure the _world_map used in c if the q0 variant.
            world_map_c = _world_map_q0;
            
            for (ray_dir_index_offset = -NR_OF_ANGLES/12; ray_dir_index_offset < NR_OF_ANGLES/12; ray_dir_index_offset++) {

                if (player_dir_index + ray_dir_index_offset < 0) {
                    // Important: the modulo in c is a *remained*-function. So it doesn't work like modulo for *negative* numbers! Therefore we have to add NR_OF_ANGLES first.
                    ray_dir_index = (player_dir_index + ray_dir_index_offset + NR_OF_ANGLES) % NR_OF_ANGLES;
                }
                else {
                    ray_dir_index = (player_dir_index + ray_dir_index_offset) % NR_OF_ANGLES;
                }
                if (use_dda) {
                    if (do_print) {
                        printf("player_dir_index: %d\n",  player_dir_index);
                    }

                    normalized_distance_to_wall = get_distance_to_wall_dda(player_pos_x, player_pos_y, ray_dir_index, sin_player_dir, cos_player_dir, do_print);
                }
                else {
                    distance_to_wall = get_distance_to_wall(player_pos_x, player_pos_y, ray_dir_index);
                    
                    // We take the angle between the player direction and the ray direction
                    ray_vs_player_angle_diff = ray_dir_index_offset;
                    // We want the absolute value of it (since our cosine only works on positive indexes)
                    if (ray_vs_player_angle_diff < 0) {
                        ray_vs_player_angle_diff = -ray_vs_player_angle_diff;
                    }
                    
                    // We take the cosine of the difference in angle to get the distance from the *plane* of the player (parallel to the camera) and the wall
                    normalized_distance_to_wall = multiply_with_fraction(distance_to_wall, cosine(ray_vs_player_angle_diff));
                }
                
                // TODO: This number is from experimentation, maybe there is a nicer way of defining it
                wall_height = WALL_HEIGHT_NUMERATOR / (normalized_distance_to_wall / 8); // This is square!

                if (!use_textures) {
                    // Using sine now, kinda works nicely
                    // TODO: make this better: 
                    light_offset = 3 + sine(wall_height*(NR_OF_ANGLES/4)/1024) / 8;
                    //light_offset = 1 + wall_height / 16;
                    if (light_offset > 15) {
                        light_offset = 15;
                    }
                    // TODO: we are using tile_value (=global) here. This is a quick-and-dirty global to get the value of a tile that was just hit
                    bar_color = tile_value * 16 + light_offset;
                    if (show_test_ray && test_ray_dir_index == ray_dir_index) {
                        bar_color = 1;
                    }
                }
                
                bar_x_start = BORDER_WIDTH + RENDER_WIDTH/2 - 1 - ray_dir_index_offset;
                
                if (use_textures) {
                    // FIXME: just testing for textures, using bar_color for the point on the wall for now
                    bar_color = hit_point_on_wall;
                }
                draw_vertical_bar(bar_x_start, wall_height, bar_color);

            }
            
        }
        */
        
        //enable_display_layer(1);
        printf("ticks %d - %d = %d\n", _ticks, start_ticks, _ticks  - start_ticks);
        //while(1);

        joystick0_state = joy_read(0);
        if(joystick0_state == 255) {
            joystick0_state = 0;   // TODO: we regard 255 to be a "starting-state", so no nothing is happpening (is this correct?)
        }
        
        key_pressed = 0;
        if(kbhit()) {
            // FIXME: im not sure this is going to work in combination with joystick state!
            key_pressed = cgetc();
            /*
            // FIXME: the code below seems to delay everything somehow...
            
            // NOTE: when a key was pressed very shortly, we might have missed it during the joy_read. But we can still register it as a key-hit. 
            if (key_pressed == 29) { // Right arrow
                joystick0_state |= 0x01;
            }
            else if (key_pressed == 157) { // Left arrow
                joystick0_state |= 0x02;
            }
            else if (key_pressed == 145) { // Up arrow
                joystick0_state |= 0x08;
            }
            else if (key_pressed == 17) { // Down arrow
                joystick0_state |= 0x04;
            }
            */
        }
        
        // Move left (strafe) -> Alt key pressed + Left arrow
        if ((joystick0_state & 0x42) == 0x42) {
        //if (key_pressed == 'a') {
            x_move = cosine((player_dir_index + NR_OF_ANGLES/4) % NR_OF_ANGLES) / 3; 
            y_move = sine((player_dir_index + NR_OF_ANGLES/4) % NR_OF_ANGLES) / 3;
            if (position_is_walkable(player_pos_x + x_move, player_pos_y + y_move)) {
                player_pos_x += x_move;
                player_pos_y += y_move;
                if (debug_is_enabled) {
                    key_pressed = 'p';
                }
            }
        }
        // Turn left -> only left arrow pressed
        else if ((joystick0_state & 0x02) == 0x02) {
        // if (key_pressed == 'q') {
            player_dir_index += NR_OF_ANGLES/64;
            if (player_dir_index >= NR_OF_ANGLES) {
                player_dir_index -= NR_OF_ANGLES;
            }
            if (debug_is_enabled) {
                key_pressed = 'p';
            }
        }
        
        // Move right (strafe) -> Alt key pressed + Right arrow
        if ((joystick0_state & 0x41) == 0x41) {
        //if (key_pressed == 'd') {
            x_move = cosine((player_dir_index + NR_OF_ANGLES - NR_OF_ANGLES/4) % NR_OF_ANGLES) / 2; 
            y_move = sine((player_dir_index + NR_OF_ANGLES - NR_OF_ANGLES/4) % NR_OF_ANGLES) / 2;
            if (position_is_walkable(player_pos_x + x_move, player_pos_y + y_move)) {
                player_pos_x += x_move;
                player_pos_y += y_move;
                if (debug_is_enabled) {
                    key_pressed = 'p';
                }
            }
        }
        // Turn right -> only right arrow pressed
        else if ((joystick0_state & 0x01) == 0x01) {
        // if (key_pressed == 'e') {
            player_dir_index -= NR_OF_ANGLES/64;
            if (player_dir_index < 0) {
                player_dir_index += NR_OF_ANGLES;
            }
            if (debug_is_enabled) {
                key_pressed = 'p';
            }
        }
        
        // Move forward
        if ((joystick0_state & 0x08) == 0x08) {
        //if (key_pressed == 'w') {
            x_move = cosine(player_dir_index % NR_OF_ANGLES) / 3; 
            y_move = sine(player_dir_index % NR_OF_ANGLES) / 3;
            if (position_is_walkable(player_pos_x + x_move, player_pos_y + y_move)) {
                player_pos_x += x_move;
                player_pos_y += y_move;
                if (debug_is_enabled) {
                    key_pressed = 'p';
                }
            }
        }
        // Move backwards
        if ((joystick0_state & 0x04) == 0x04) {
        // if (key_pressed == 's') {
            x_move = cosine(player_dir_index % NR_OF_ANGLES) / 3; 
            y_move = sine(player_dir_index % NR_OF_ANGLES) / 3;
            if (position_is_walkable(player_pos_x - x_move, player_pos_y - y_move)) {
                player_pos_x -= x_move;
                player_pos_y -= y_move;
                if (debug_is_enabled) {
                    key_pressed = 'p';
                }
            }
        }
        
        
        // Debug: Draw ray-carting into to log screen
        if (key_pressed == 'p') {
            debug_is_enabled = 1;
            enable_display_layer(1);
            printf("%c", 147);
/*
            sin_player_dir = sine(player_dir_index);
            cos_player_dir = cosine(player_dir_index);
            
            if (1) {
                printf("player_dir_index: %d\n",  player_dir_index);
            }
            normalized_distance_to_wall = get_distance_to_wall_dda(player_pos_x, player_pos_y, test_ray_dir_index, sin_player_dir, cos_player_dir, 1);
            if (1) {
                printf("normalized_distance_to_wall: %d\n",  normalized_distance_to_wall);
            }
            */
        }
        // Debug: Turn off log screen
        if (key_pressed == 'o') {
            debug_is_enabled = 0;
            disable_display_layer(1);
        }
        /*
        if (key_pressed == 'u') {
            enable_display_layer(1);
        }
        */
        // Debug: Turn on/off test ray
        if (key_pressed == 't') {
            show_test_ray = 1 - show_test_ray;
        }
        // Debug: Turn on/off textures
        if (key_pressed == 'x') {
            use_textures = 1 - use_textures;
        }
        // Debug: rotate test ray to the left
        if (key_pressed == 'j') {
            test_ray_dir_index += 1;
            if (debug_is_enabled) {
                key_pressed = 'p';
            }
        }
        // Debug: rotate test ray to the right
        if (key_pressed == 'l') {
            test_ray_dir_index -= 1;
            if (debug_is_enabled) {
                key_pressed = 'p';
            }
        }
        
    }
}

/*
void print_start_ray(u8 x_tile, u8 y_tile, u8 dx, u8 dy, i16 x_step, i16 y_step, i8 tile_step_x, i8 tile_step_y, i16 x_intercept, i16 y_intercept, u16 ray_dir_index) {
    printf("x_tile+dx: %d+%d\n", x_tile, dx);
    printf("y_tile+dy: %d+%d\n", y_tile, dy);

    if (x_step >= 0) {
        printf("x_step: %d+%d\n", x_step/256, x_step%256);
    }
    else {
        printf("x_step: -%d+%d\n", (-x_step)/256, (-x_step)%256);
    }
    if (y_step >= 0) {
        printf("y_step: %d+%d\n", y_step/256, y_step%256);
    }
    else {
        printf("y_step: -%d+%d\n", (-y_step)/256, (-y_step)%256);
    }

    printf("tile_step_x: %d\n", tile_step_x);
    printf("tile_step_y: %d\n", tile_step_y);
    printf("ray_dir_index: %d\n",  ray_dir_index);
    printf("\n");

    printf("x_tile, y_tile: %d, %d\n", x_tile, y_tile);
    printf("x_intercept: %d+%d\n", x_intercept/256, x_intercept%256);
    printf("y_intercept: %d+%d\n", y_intercept/256, y_intercept%256);
}
*/


u16 get_distance_to_wall_dda_asm(u16 player_pos_x, u16 player_pos_y, i16 ray_dir_index, i16 sin_player_dir, i16 cos_player_dir) {
    u8 x_tile, y_tile;
    u8 x_tile_orig, y_tile_orig;
    u8 dx, dy;
    u16 distance_to_wall = 0;

    // FIXME: temporary variable to use dda in c
    u8 use_asm_dda = 1;
    
    // FIXME: we are ignoring this warning for x_tile and y_tile right now.
    // WANRING: y_intercept and x_intercept can become negative (outside of map). This is why we make x_tile, y_tile, y_intercept and x_intercept signed numbers.
    /*    
    printf("sin_player_dir: %d\n", sin_player_dir);
    printf("cos_player_dir: %d\n", cos_player_dir);
    printf("ray_dir_index: %d\n",  ray_dir_index);
    */    
    x_tile = player_pos_x >> 8; // high byte  // FIXME: this can directly be put into _x_tile
    dx = player_pos_x & 0xFF;    // low byte  // FIXME: this can directly be put into _dx
    x_tile_orig = x_tile;
    
    y_tile = player_pos_y >> 8; // high byte  // FIXME: this can directly be put into _y_tile
    dy = player_pos_y & 0xFF;    // low byte  // FIXME: this can directly be put into _dy
    y_tile_orig = y_tile;

 
    if (ray_dir_index >= 0 && ray_dir_index <= NR_OF_ANGLES / 4) { // North East
    
        _world_map = (u16)&_world_map_q0;
        world_map_c = _world_map_q0;
    
        _x_tile = x_tile;
        _y_tile = y_tile;
        _minus_dx = 255-dx;
        _minus_dy = 255-dy;
        _ray_dir_index = ray_dir_index;
    
        _player_pos_x_local = player_pos_x;
        _player_pos_y_local = player_pos_y;
        _sin_player_dir_local = sin_player_dir;
        _cos_player_dir_local = cos_player_dir;

        if (use_asm_dda) {
            _prepare_single_raycast();
            _cast_ray_dda_qdrt0();
            /*
            enable_display_layer(1);
            printf("sin_player_dir: %d\n", sin_player_dir);
            printf("cos_player_dir: %d\n", cos_player_dir);
            printf("ray_dir_index: %d\n",  ray_dir_index);
            printf("x_tile+dx: %d+%d\n", x_tile, dx);
            printf("y_tile+dy: %d+%d\n", y_tile, dy);
            printf("ray_dir_index: %d\n",  ray_dir_index);
            printf("_ray_dir_index: %d\n",  _ray_dir_index);
            printf("\n");
            if (_x_step >= 0) {
                printf("_x_step: %d+%d\n", _x_step/256, _x_step%256);
            }
            else {
                printf("_x_step: -%d+%d\n", (-_x_step)/256, (-_x_step)%256);
            }
            if (_y_step >= 0) {
                printf("_y_step: %d+%d\n", _y_step/256, _y_step%256);
            }
            else {
                printf("_y_step: -%d+%d\n", (-_y_step)/256, (-_y_step)%256);
            }
            printf("_x_intercept: %d+%d\n", _x_intercept/256, _x_intercept%256);
            printf("_y_intercept: %d+%d\n", _y_intercept/256, _y_intercept%256);
            printf("_tile_value: %d\n", _tile_value);
            printf("_total_delta_x: %d+%d\n", _total_delta_x/256, _total_delta_x%256);
            printf("_total_delta_y: %d+%d\n", _total_delta_y/256, _total_delta_y%256);
            printf("_distance_to_wall: %d+%d\n", _distance_to_wall/256, _distance_to_wall%256);
            printf("_hit_point_on_wall: %d\n", _hit_point_on_wall);
            while(1);
            */
            distance_to_wall = _distance_to_wall;
            tile_value = _tile_value;
            hit_point_on_wall = _hit_point_on_wall;
        }
        else {
            // do_print = 1;
//            distance_to_wall = get_distance_to_wall_dda(_player_pos_x_local, _player_pos_y_local, _ray_dir_index, _sin_player_dir_local, _cos_player_dir_local, do_print);
            // enable_display_layer(1);
            // while(1);
        }
        
    }
    else if (ray_dir_index > NR_OF_ANGLES / 4 && ray_dir_index <= NR_OF_ANGLES / 2) { // North West
    
        _world_map = (u16)&_world_map_q1;
        world_map_c = _world_map_q1;

        _x_tile = MAP_WIDTH_MAX-1-x_tile;
        _y_tile = y_tile;
        _minus_dx = dx;
        _minus_dy = 255-dy;
        // FIXME: should we substract 1 here?        
        _ray_dir_index = NR_OF_ANGLES/2-ray_dir_index;
    
        _player_pos_x_local = _x_tile*256 + 255-dx;
        _player_pos_y_local = player_pos_y;
        _sin_player_dir_local = sin_player_dir;
        // FIXME: can the asm deal with -256?? (in sin or cos) or will this never be the case/a problem?
        _cos_player_dir_local = -cos_player_dir;

        if (use_asm_dda) {
            _prepare_single_raycast();
            _cast_ray_dda_qdrt0();
            distance_to_wall = _distance_to_wall;
            tile_value = _tile_value;
            if (_hit_vertical_wall) {
                hit_point_on_wall = _hit_point_on_wall;
            }
            else {
                hit_point_on_wall = 255-_hit_point_on_wall;
            }
        }
        else {
//            distance_to_wall = get_distance_to_wall_dda(_player_pos_x_local, _player_pos_y_local, _ray_dir_index, _sin_player_dir_local, _cos_player_dir_local, do_print);
        }
    
    }
    else if (ray_dir_index > NR_OF_ANGLES / 2 && ray_dir_index < (NR_OF_ANGLES / 4) * 3) { // South West
        _world_map = (u16)&_world_map_q2;
        world_map_c = _world_map_q2;

        _x_tile = MAP_WIDTH_MAX-1-x_tile;
        _y_tile = MAP_HEIGHT_MAX-1-y_tile;
        _minus_dx = dx;
        _minus_dy = dy;
        _ray_dir_index = ray_dir_index-NR_OF_ANGLES/2;
    
        _player_pos_x_local = _x_tile*256 + 255-dx;
        _player_pos_y_local = _y_tile*256 + 255-dy;
        // FIXME: can the asm deal with -256?? (in sin or cos) or will this never be the case/a problem?
        _sin_player_dir_local = -sin_player_dir;
        _cos_player_dir_local = -cos_player_dir;
        if (use_asm_dda) {
            _prepare_single_raycast();
            _cast_ray_dda_qdrt0();
            distance_to_wall = _distance_to_wall;
            tile_value = _tile_value;
            hit_point_on_wall = 255-_hit_point_on_wall;
        }
        else {
//            distance_to_wall = get_distance_to_wall_dda(_player_pos_x_local, _player_pos_y_local, _ray_dir_index, _sin_player_dir_local, _cos_player_dir_local, do_print);
        }
    
    }
    else {
        _world_map = (u16)&_world_map_q3;
        world_map_c = _world_map_q3;

        _x_tile = x_tile;
        _y_tile = MAP_HEIGHT_MAX-1-y_tile;
        _minus_dx = 255-dx;
        _minus_dy = dy;
        // FIXME: should we substract 1 here?        
        _ray_dir_index = NR_OF_ANGLES-ray_dir_index;
    
        _player_pos_x_local = _x_tile*256 + dx;
        _player_pos_y_local = _y_tile*256 + 255-dy;
        // FIXME: can the asm deal with -256?? (in sin or cos) or will this never be the case/a problem?
        _sin_player_dir_local = -sin_player_dir;
        _cos_player_dir_local = cos_player_dir;
    
        if (use_asm_dda) {
            _prepare_single_raycast();
            _cast_ray_dda_qdrt0();
            distance_to_wall = _distance_to_wall;
            tile_value = _tile_value;
            if (_hit_vertical_wall) {
                hit_point_on_wall = 255-_hit_point_on_wall;
            }
            else {
                hit_point_on_wall = _hit_point_on_wall;
            }
        }
        else {
//            distance_to_wall = get_distance_to_wall_dda(_player_pos_x_local, _player_pos_y_local, _ray_dir_index, _sin_player_dir_local, _cos_player_dir_local, do_print);
        }

    }
    
    return distance_to_wall;
    
}

/*
u16 get_distance_to_wall_dda(u16 player_pos_x, u16 player_pos_y, i16 ray_dir_index, i16 sin_player_dir, i16 cos_player_dir, u8 do_print) {
    u8 x_tile, y_tile;
    u8 x_tile_orig, y_tile_orig;
    u8 dx, dy;
    i8 tile_step_x, tile_step_y;
    i16 x_step, y_step;
    i16 x_intercept, y_intercept;
    i16 total_delta_x, total_delta_y;
    u8 making_progress;
    u8 travelled_in_y, travelled_in_x;
    u16 distance_to_wall;

    // WANRING: y_intercept and x_intercept can become negative (outside of map). This is why we make x, y, y_intercept and x_intercept signed numbers.
    
    if (do_print) {
        printf("sin_player_dir: %d\n", sin_player_dir);
        printf("cos_player_dir: %d\n", cos_player_dir);
        printf("ray_dir_index: %d\n",  ray_dir_index);
    }
    
    x_tile = player_pos_x >> 8; // high byte
    dx = player_pos_x & 0xFF;    // low byte
    x_tile_orig = x_tile;
    
    y_tile = player_pos_y >> 8; // high byte
    dy = player_pos_y & 0xFF;    // low byte
    y_tile_orig = y_tile;

    distance_to_wall = 0;
    making_progress = 1;
    travelled_in_y = 0;
    travelled_in_x = 0;
    
    if (ray_dir_index >= 0 && ray_dir_index <= NR_OF_ANGLES / 4) { // North East
        tile_step_x = 1;
        tile_step_y = 1;
        
        x_step = invtangent(ray_dir_index);
        y_step = tangent(ray_dir_index);
        
        x_intercept = player_pos_x + multiply_with_fraction(x_step, 255-dy);
        y_intercept = player_pos_y + multiply_with_fraction(y_step, 255-dx);
        
        if (do_print) {
            print_start_ray(x_tile, y_tile, dx, dy, x_step, y_step, tile_step_x, tile_step_y, x_intercept, y_intercept, ray_dir_index);
        }
        
        while(making_progress) {
            making_progress = 0;
            // Travel in x
            travelled_in_x = 0;
            while (travelled_in_y || y_intercept <= (y_tile+1)*256) { // As long as my yIntercept hasn't become bigger (=outside my y-row) than the y I am in (because then we have to move x)
                travelled_in_y = 0;
                tile_value = world_map_c[y_tile*MAP_WIDTH_MAX+x_tile+tile_step_x];
                if (tile_value) {
                    total_delta_x = (x_tile+tile_step_x)*256 - player_pos_x;
                    total_delta_y = y_intercept - player_pos_y;
                    hit_point_on_wall = 255-y_intercept%256;
                    distance_to_wall = multiply_with_fraction(total_delta_x, cos_player_dir) + multiply_with_fraction(total_delta_y, sin_player_dir);
                    if (do_print) {
                        printf("hit!\n");
                        printf("hit vertical wall at: %d, %d\n", x_tile+tile_step_x, y_tile);
                        printf("hit point: %d, %d+%d\n", x_tile+tile_step_x, y_intercept/256, y_intercept%256);
                        printf("total_delta_x, total_delta_y: %d, %d\n", total_delta_x, total_delta_y);
                        // TODO: added this one, either remove it or add to all of them
                        printf("hit_point_on_wall: %d\n", hit_point_on_wall);
                    }
                    goto we_hit_the_wall;
                }
                x_tile += tile_step_x; // We step to the next x-tile
                y_intercept += y_step; // the next intercept with the vertical bar will be y_step incremented
                making_progress = 1;
                travelled_in_x = 1;
                if (do_print) {
                    printf("stepping to next x tile: %d, %d\n", x_tile, y_tile);
                    printf("x_intercept: %d+%d\n", x_intercept/256, x_intercept%256);
                    printf("y_intercept: %d+%d\n", y_intercept/256, y_intercept%256);
                }
            }
            // Travel in y
            travelled_in_y = 0;
            while (travelled_in_x || x_intercept <= (x_tile+1)*256) { // As long as my xIntercept hasn't become bigger (outside my x-column) than the x I am in (because then we have to move x!)
                travelled_in_x = 0;                
                tile_value = world_map_c[(y_tile+tile_step_y)*MAP_WIDTH_MAX+x_tile];
                if (tile_value) {
                    total_delta_x = x_intercept - player_pos_x;
                    total_delta_y = (y_tile+tile_step_y)*256 - player_pos_y;
                    hit_point_on_wall = x_intercept%256;
                    distance_to_wall = multiply_with_fraction(total_delta_x, cos_player_dir) + multiply_with_fraction(total_delta_y, sin_player_dir);
                    if (do_print) {
                        printf("hit horizontal wall at: %d, %d\n", x_tile, y_tile+tile_step_y);
                        printf("hit point: %d+%d, %d\n", x_intercept/256, x_intercept%256, y_tile);
                        printf("total_delta_x, total_delta_y: %d, %d\n", total_delta_x, total_delta_y);
                        // TODO: added this one, either remove it or add to all of them
                        printf("hit_point_on_wall: %d\n", hit_point_on_wall);
                    }
                    goto we_hit_the_wall;
                }
                y_tile += tile_step_y; // We step to the next y-tile
                x_intercept += x_step; // the next intercept with the horizontal bar will be x_step incremented
                making_progress = 1;
                travelled_in_y = 1;
                if (do_print) {
                    printf("stepping to next y tile: %d, %d\n", x_tile, y_tile);
                    printf("x_intercept: %d+%d\n", x_intercept/256, x_intercept%256);
                    printf("y_intercept: %d+%d\n", y_intercept/256, y_intercept%256);
                }
            }
        }
    }
    else if (ray_dir_index > NR_OF_ANGLES / 4 && ray_dir_index <= NR_OF_ANGLES / 2) { // North West
        tile_step_x = -1;
        tile_step_y = 1;

        x_step = invtangent(ray_dir_index);
        y_step = -tangent(ray_dir_index); // TODO: why should this be negative while the North East one should not be negated (even though the both move in the same y-direction) -> maybe because of tangent() being the opposite when angle is > 90 degrees
        
        x_intercept = player_pos_x + multiply_with_fraction(x_step, 255-dy);
        y_intercept = player_pos_y + multiply_with_fraction(y_step, dx);

        if (do_print) {
            print_start_ray(x_tile, y_tile, dx, dy, x_step, y_step, tile_step_x, tile_step_y, x_intercept, y_intercept, ray_dir_index);
        }
        //    PROBLEM: when y_intercept == 0+11 (which is for example *rounded down* from 0+11.3)
        //            you may end up adding y_intercept many times, but hitting the x-tile too quickly
        //            this will end up in an infinite loop.
        //         
        //    SOLUTION: if you just decided that you exhausted in one direction (and did not hit a wall), always move at least once in the other direction)
        while(making_progress) {
            making_progress = 0;
            // Travel in x
            travelled_in_x = 0;
            while (travelled_in_y || y_intercept <= (y_tile+1)*256) { // As long as my yIntercept hasn't become bigger (=outside my y-row) than the y I am in (because then we have to move y)
                travelled_in_y = 0;
                tile_value = world_map_c[y_tile*MAP_WIDTH_MAX+x_tile+tile_step_x];
                if (tile_value) {
                    total_delta_x = x_tile*256 - player_pos_x;
                    total_delta_y = y_intercept - player_pos_y;
                    if (do_print) {
                        printf("hit!\n");
                        printf("hit vertical wall at: %d, %d\n", x_tile+tile_step_x, y_tile);
                        printf("hit point: %d, %d+%d\n", x_tile+tile_step_x, y_intercept/256, y_intercept%256);
                        printf("total_delta_x, total_delta_y: %d, %d\n", total_delta_x, total_delta_y);
                    }
                    hit_point_on_wall = y_intercept%256;
                    distance_to_wall = multiply_with_fraction(total_delta_x, cos_player_dir) + multiply_with_fraction(total_delta_y, sin_player_dir);
                    goto we_hit_the_wall;
                }
                x_tile += tile_step_x; // We step to the next x-tile
                y_intercept += y_step; // the next intercept with the vertical bar will be y_step incremented
                making_progress = 1;
                travelled_in_x = 1;
                if (do_print) {
                    printf("stepping to next x tile: %d, %d\n", x_tile, y_tile);
                    printf("x_intercept: %d+%d\n", x_intercept/256, x_intercept%256);
                    printf("y_intercept: %d+%d\n", y_intercept/256, y_intercept%256);
                }
            }
            // Travel in y
            travelled_in_y = 0;
            while (travelled_in_x || x_intercept >= x_tile*256) { // As long as my xIntercept hasn't become bigger (outside my x-column) than the x I am in (because then we have to move y one lower!)
                travelled_in_x = 0;                
                tile_value = world_map_c[(y_tile+tile_step_y)*MAP_WIDTH_MAX+x_tile];
                if (tile_value) {
                    total_delta_x = x_intercept - player_pos_x;
                    total_delta_y = (y_tile + tile_step_y)*256 - player_pos_y;
                    if (do_print) {
                        printf("hit horizontal wall at: %d, %d\n", x_tile, y_tile+tile_step_y);
                        printf("hit point: %d+%d, %d\n", x_intercept/256, x_intercept%256, y_tile);
                        printf("total_delta_x, total_delta_y: %d, %d\n", total_delta_x, total_delta_y);
                    }
                    hit_point_on_wall = x_intercept%256;
                    distance_to_wall = multiply_with_fraction(total_delta_x, cos_player_dir) + multiply_with_fraction(total_delta_y, sin_player_dir);
                    goto we_hit_the_wall;
                }
                y_tile += tile_step_y; // We step to the next y-tile
                x_intercept += x_step; // the next intercept with the horizontal bar will be x_step incremented
                making_progress = 1;
                travelled_in_y = 1;
                if (do_print) {
                    printf("stepping to next y tile: %d, %d\n", x_tile, y_tile);
                    printf("x_intercept: %d+%d\n", x_intercept/256, x_intercept%256);
                    printf("y_intercept: %d+%d\n", y_intercept/256, y_intercept%256);
                }
            }
        }
    }
    // Note: this is NOT <= 3/4 * NR_OF_ANGLES, because for ray_dir_index = 240 the tan is positive, yet for all others (160-239) it is negative. Since we negate it, we can't also deal with 240 here.
    else if (ray_dir_index > NR_OF_ANGLES / 2 && ray_dir_index < (NR_OF_ANGLES / 4) * 3) { // South West
        tile_step_x = -1;
        tile_step_y = -1;
        
        x_step = -invtangent(ray_dir_index);// tile_step_x = -1, so x_step should be negative. But direction causes invtan to be positive, so we should negate.
        y_step = -tangent(ray_dir_index); // tile_step_y = -1 so y_step should be negative. But direction causes tan to be positive, so we should negate.
        
        x_intercept = player_pos_x + multiply_with_fraction(x_step, dy);
        y_intercept = player_pos_y + multiply_with_fraction(y_step, dx);

        if (do_print) {
            print_start_ray(x_tile, y_tile, dx, dy, x_step, y_step, tile_step_x, tile_step_y, x_intercept, y_intercept, ray_dir_index);
        }
            
        while(making_progress) {
            making_progress = 0;
            // Travel in x
            travelled_in_x = 0;
            while (travelled_in_y || y_intercept >= y_tile*256) { // As long as my yIntercept hasn't become smaller (=outside my y-row) than the y I am in (because then we have to move y)
                travelled_in_y = 0;
                tile_value = world_map_c[y_tile*MAP_WIDTH_MAX+x_tile+tile_step_x];
                if (tile_value) {
                    total_delta_x = x_tile*256 - player_pos_x;
                    total_delta_y = y_intercept - player_pos_y;
                    if (do_print) {
                        printf("hit!\n");
                        printf("hit vertical wall at: %d, %d\n", x_tile+tile_step_x, y_tile);
                        printf("hit point: %d, %d+%d\n", x_tile+tile_step_x, y_intercept/256, y_intercept%256);
                        printf("total_delta_x, total_delta_y: %d, %d\n", total_delta_x, total_delta_y);
                    }
                    hit_point_on_wall = y_intercept%256;
                    distance_to_wall = multiply_with_fraction(total_delta_x, cos_player_dir) + multiply_with_fraction(total_delta_y, sin_player_dir);
                    goto we_hit_the_wall;
                }
                x_tile += tile_step_x; // We step to the next x-tile
                y_intercept += y_step; // the next intercept with the vertical bar will be y_step incremented
                making_progress = 1;
                travelled_in_x = 1;
                if (do_print) {
                    printf("stepping to next x tile: %d, %d\n", x_tile, y_tile);
                    printf("x_intercept: %d+%d\n", x_intercept/256, x_intercept%256);
                    printf("y_intercept: %d+%d\n", y_intercept/256, y_intercept%256);
                }
            }
            // Travel in y
            travelled_in_y = 0;
            while (travelled_in_x || x_intercept >= x_tile*256) { // As long as my xIntercept hasn't become bigger (outside my x-column) than the x I am in (because then we have to move y one lower!)
                travelled_in_x = 0;
                tile_value = world_map_c[(y_tile+tile_step_y)*MAP_WIDTH_MAX+x_tile];
                if (tile_value) {
                    total_delta_x = x_intercept - player_pos_x;
                    total_delta_y = y_tile*256 - player_pos_y;
                    if (do_print) {
                        printf("hit horizontal wall at: %d, %d\n", x_tile, y_tile+tile_step_y);
                        printf("hit point: %d+%d, %d\n", x_intercept/256, x_intercept%256, y_tile);
                        printf("total_delta_x, total_delta_y: %d, %d\n", total_delta_x, total_delta_y);
                    }
                    hit_point_on_wall = 255-x_intercept%256;
                    distance_to_wall = multiply_with_fraction(total_delta_x, cos_player_dir) + multiply_with_fraction(total_delta_y, sin_player_dir);
                    goto we_hit_the_wall;
                }
                y_tile += tile_step_y; // We step to the next y-tile
                x_intercept += x_step; // the next intercept with the horizontal bar will be x_step incremented
                making_progress = 1;
                travelled_in_y = 1;
                if (do_print) {
                    printf("stepping to next y tile: %d, %d\n", x_tile, y_tile);
                    printf("x_intercept: %d+%d\n", x_intercept/256, x_intercept%256);
                    printf("y_intercept: %d+%d\n", y_intercept/256, y_intercept%256);
                }
            }
        }
    }
    else {  // South East
        tile_step_x = 1;
        tile_step_y = -1;
        
        x_step = -invtangent(ray_dir_index); // tile_step_x = 1, so x_step should be positivee. But direction causes invtan to be negative, so we should negate.
        y_step = tangent(ray_dir_index);
        
        x_intercept = player_pos_x + multiply_with_fraction(x_step, dy);
        y_intercept = player_pos_y + multiply_with_fraction(y_step, 255-dx);
        
        if (do_print) {
            print_start_ray(x_tile, y_tile, dx, dy, x_step, y_step, tile_step_x, tile_step_y, x_intercept, y_intercept, ray_dir_index);
        }
        
        while(making_progress) {
            making_progress = 0;
            // Travel in x
            travelled_in_x = 0;
            while (travelled_in_y || y_intercept >= y_tile*256) { // As long as my yIntercept hasn't become smaller (=outside my y-row) than the y I am in (because then we have to move x)
                travelled_in_y = 0;
                tile_value = world_map_c[y_tile*MAP_WIDTH_MAX+x_tile+tile_step_x];
                if (tile_value) {
                    total_delta_x = (x_tile + tile_step_x)*256 - player_pos_x;
                    total_delta_y = y_intercept - player_pos_y;
                    if (do_print) {
                        printf("hit!\n");
                        printf("hit vertical wall at: %d, %d\n", x_tile+tile_step_x, y_tile);
                        printf("hit point: %d, %d+%d\n", x_tile+tile_step_x, y_intercept/256, y_intercept%256);
                        printf("total_delta_x, total_delta_y: %d, %d\n", total_delta_x, total_delta_y);
                    }
                    hit_point_on_wall = 255-y_intercept%256;
                    distance_to_wall = multiply_with_fraction(total_delta_x, cos_player_dir) + multiply_with_fraction(total_delta_y, sin_player_dir);
                    goto we_hit_the_wall;
                }
                x_tile += tile_step_x; // We step to the next x-tile
                y_intercept += y_step; // the next intercept with the vertical bar will be y_step incremented
                making_progress = 1;
                travelled_in_x = 1;
                if (do_print) {
                    printf("stepping to next x tile: %d, %d\n", x_tile, y_tile);
                    printf("x_intercept: %d+%d\n", x_intercept/256, x_intercept%256);
                    printf("y_intercept: %d+%d\n", y_intercept/256, y_intercept%256);
                }
            }
            // Travel in y
            travelled_in_y = 0;
            while (travelled_in_x || x_intercept <= (x_tile+1)*256) { // As long as my xIntercept hasn't become bigger (outside my x-column) than the x I am in (because then we have to move x!)
                travelled_in_x = 0;
                tile_value = world_map_c[(y_tile+tile_step_y)*MAP_WIDTH_MAX+x_tile];
                if (tile_value) {
                    total_delta_x = x_intercept - player_pos_x;
                    total_delta_y = y_tile*256 - player_pos_y;
                    if (do_print) {
                        printf("hit horizontal wall at: %d, %d\n", x_tile, y_tile+tile_step_y);
                        printf("hit point: %d+%d, %d\n", x_intercept/256, x_intercept%256, y_tile);
                        printf("total_delta_x, total_delta_y: %d, %d\n", total_delta_x, total_delta_y);
                    }
                    hit_point_on_wall = 255-x_intercept%256;
                    distance_to_wall = multiply_with_fraction(total_delta_x, cos_player_dir) + multiply_with_fraction(total_delta_y, sin_player_dir);
                    goto we_hit_the_wall;
                }
                y_tile += tile_step_y; // We step to the next y-tile
                x_intercept += x_step; // the next intercept with the horizontal bar will be x_step incremented
                making_progress = 1;
                travelled_in_y = 1;
                if (do_print) {
                    printf("stepping to next y tile: %d, %d\n", x_tile, y_tile);
                    printf("x_intercept: %d+%d\n", x_intercept/256, x_intercept%256);
                    printf("y_intercept: %d+%d\n", y_intercept/256, y_intercept%256);
                }
            }
        }
    }
    

    if (!making_progress) {
        // FIXME: ulgy
        enable_display_layer(1);
        printf("not making progress!!\n");
        
        printf("tile_step_x: %d\n", tile_step_x);
        printf("tile_step_y: %d\n", tile_step_y);
        printf("ray_dir_index: %d\n",  ray_dir_index);
        printf("\n");
        
        if (x_step >= 0) {
            printf("x_step: %d+%d\n", x_step/256, x_step%256);
        }
        else {
            printf("x_step: -%d+%d\n", -x_step/256, -x_step%256);
        }
        if (y_step >= 0) {
            printf("y_step: %d+%d\n", y_step/256, y_step%256);
        }
        else {
            printf("y_step: -%d+%d\n", (-y_step)/256, (-y_step)%256);
        }
        
        printf("x_tile_orig+dx: %d+%d\n", x_tile_orig, dx);
        printf("y_tile_orig+dy: %d+%d\n", y_tile_orig, dy);
        printf("x_intercept: %d+%d\n", x_intercept/256, x_intercept%256);
        printf("y_intercept: %d+%d\n", y_intercept/256, y_intercept%256);
        while(1);
    }
    
we_hit_the_wall:

    return distance_to_wall;
}
*/

/*
u16 get_distance_to_wall(u16 player_pos_x, u16 player_pos_y, u16 ray_dir_index) {
    u16 distance_to_wall = 0;
    u8 keep_searching = 1;
    
    u16 ray_point_x;
    u16 ray_point_y;
    i16 steps_times_angle_x;
    i16 steps_times_angle_y;
    u16 nr_of_ray_steps = 0;
    
    i16 x_part_of_angle = cosine(ray_dir_index); // this is the x-part of the 256-length vector
    i16 y_part_of_angle = sine(ray_dir_index); // this is the y-part of the 256-length vector
    
    // We do a 'naive' search here: we keep increasing the length of our ray with small increments until we hit a wall.
    // Wolfenstein 3D did this in a much more efficient way, but this is still a todo.
    while (keep_searching) {
        
        // FIXME: when the number of steps exceeds 128 (or 256) we reach the limit of i16 (of the multiplication). We should separate all full-256 length pieces.
        steps_times_angle_x = (nr_of_ray_steps * x_part_of_angle);
        steps_times_angle_y = (nr_of_ray_steps * y_part_of_angle);
        
        ray_point_x = player_pos_x + steps_times_angle_x / SEARCH_STEP_FRACTION; // each step is 1/16th of the x-part of the 256-length vector
        ray_point_y = player_pos_y + steps_times_angle_y / SEARCH_STEP_FRACTION; // each step is 1/16th of the y-part of the 256 length vector
    
        // TODO: check if ray_point is not out-of-bound
        
        if (position_is_wall(ray_point_x, ray_point_y)) {
            keep_searching = 0;
            break;
        }
        
        nr_of_ray_steps += 1;
        if (nr_of_ray_steps > 128) { // FIXME: see FIXME-comment above
            keep_searching = 0;
            break;
        }
    }
    
    distance_to_wall = nr_of_ray_steps * (256 / SEARCH_STEP_FRACTION);
    
    return distance_to_wall;
}
*/

u8 position_is_wall(i16 ray_point_x, i16 ray_point_y) {
    u8 wall_hit = 0;
    i16 wall_index_x = 0;
    i16 wall_index_y = 0;
    
    wall_index_x = ray_point_x / TILE_SIZE;
    wall_index_y = ray_point_y / TILE_SIZE;

    tile_value = _world_map_q0[wall_index_y*MAP_WIDTH_MAX+wall_index_x];
    if (tile_value > 0) {
        // TODO: as a workaround we return the tile_value itself (instead of a 0 or 1). There might be a better way to return this value.
        wall_hit = tile_value;
    }
    
    return wall_hit;
}

u8 position_is_walkable(i16 pos_x, i16 pos_y) {
    u8 is_walkable = 1;
    i16 margin = TILE_SIZE / 3; // In Wolf 3D it seems 37% margin (of a full block)
    
    // Center
    if (position_is_wall(pos_x, pos_y)) {
        is_walkable = 0;
    }
    // West
    if (position_is_wall(pos_x - margin, pos_y)) {
        is_walkable = 0;
    }
    // East
    if (position_is_wall(pos_x + margin, pos_y)) {
        is_walkable = 0;
    }
    // North
    if (position_is_wall(pos_x, pos_y + margin)) {
        is_walkable = 0;
    }
    // South
    if (position_is_wall(pos_x, pos_y - margin)) {
        is_walkable = 0;
    }
    // North East
    if (position_is_wall(pos_x + margin, pos_y + margin)) {
        is_walkable = 0;
    }
    // North West
    if (position_is_wall(pos_x - margin, pos_y + margin)) {
        is_walkable = 0;
    }
    // South East
    if (position_is_wall(pos_x + margin, pos_y - margin)) {
        is_walkable = 0;
    }
    // South West
    if (position_is_wall(pos_x - margin, pos_y - margin)) {
        is_walkable = 0;
    }
    
    return is_walkable;
}

void copy_world_map_to_four_quadrants() {
    i8 x, y;
    for (y = 0; y < MAP_HEIGHT; y++) { // dividing by 2 for an odd number means rounding down (so the middle row won't be copied)
        for (x = 0; x < MAP_WIDTH; x++) {
            _world_map_q0[                   y*MAP_WIDTH_MAX + x                ] = world_map[MAP_HEIGHT-1-y][x];
            _world_map_q1[                   y*MAP_WIDTH_MAX + MAP_WIDTH_MAX-1-x] = world_map[MAP_HEIGHT-1-y][x];  // q1 is x-flipped
            _world_map_q2[(MAP_HEIGHT_MAX-1-y)*MAP_WIDTH_MAX + MAP_WIDTH_MAX-1-x] = world_map[MAP_HEIGHT-1-y][x];  // q2 is x- and y-flipped
            _world_map_q3[(MAP_HEIGHT_MAX-1-y)*MAP_WIDTH_MAX + x                ] = world_map[MAP_HEIGHT-1-y][x];  // q2 is y-flipped
        }
    }
}

void create_new_palette() {
    u8 color_index, light_offset;
    
    // Grey (1)
    
    // This is already in default palette
    
    // Red  (2)
    for (light_offset = 0; light_offset < 16; light_offset++) { 
        color_index = 32 + light_offset;
        set_palette_color(color_index, light_offset, 0, 0);
    }

    // Green (3)
    for (light_offset = 0; light_offset < 16; light_offset++) { 
        color_index = 48 + light_offset;
        set_palette_color(color_index, 0, light_offset, 0);
    }
    
    // Blue (4)
    for (light_offset = 0; light_offset < 16; light_offset++) { 
        color_index = 64 + light_offset;
        set_palette_color(color_index, 0, 0, light_offset);
    }
    
    // Cyan (5)
    for (light_offset = 0; light_offset < 16; light_offset++) { 
        color_index = 80 + light_offset;
        set_palette_color(color_index, 0, light_offset, light_offset);
    }
    
    // Magenta (6)
    for (light_offset = 0; light_offset < 16; light_offset++) { 
        color_index = 96 + light_offset;
        set_palette_color(color_index, light_offset, 0, light_offset);
    }
    
    // Yellow (7)
    for (light_offset = 0; light_offset < 16; light_offset++) { 
        color_index = 112 + light_offset;
        set_palette_color(color_index, light_offset, light_offset, 0);
    }
}

/*
i16 multiply_with_fraction (i16 number, i16 fraction) {
    
    // Note: since we don't have floating point numbers, we use this special function to multiplie a 16-bit number 
    // with a fraction (like 0.83 from a cosine). The 'fraction' is actually a number between 0 and (including) 256. 
    // After the multiplication we divide by 256. This is to prevent overflow and/or to prevent loss of accuracy.
    
    u16 multiplied_number;
    u16 high_bits, low_bits;
    u16 multiplied_high_bits, multiplied_low_bits;
    u16 high_bits_of_multiplied_low_bits;
    i8 negate_result = 0;
    
    if (number < 0) {
        number = -number;
        negate_result = 1-negate_result;
    }
    if (fraction < 0) {
        fraction = -fraction;
        negate_result = 1-negate_result;
    }
    
    high_bits = number >> 8;
    low_bits = number & 0xFF;
    
    multiplied_low_bits = low_bits * fraction;
    multiplied_high_bits = high_bits * fraction;
    
    high_bits_of_multiplied_low_bits = multiplied_low_bits >> 8;
    
    multiplied_number = multiplied_high_bits + high_bits_of_multiplied_low_bits;

    if (negate_result) {
        multiplied_number = -multiplied_number;
    }
    return multiplied_number;
}
*/

// From: https://www.commanderx16.com/forum/index.php?/topic/795-cc65-cbm_k_load-into-banked-ram/

#define IGNORE_LFN              0
#define FILE_SYSTEM             8
#define SA_IGNORE_HEADER        0
#define LOAD_TO_X16_RAM         0
#define LOAD_TO_VERA_00000      1
#define LOAD_TO_VERA_10000      2

void load_file_into_banked_ram(char * file_name)
{  
   cbm_k_setnam(file_name); // NOTE: file name must be in lowercase here, but uppercase in the file system
   cbm_k_setlfs(IGNORE_LFN, FILE_SYSTEM, SA_IGNORE_HEADER);
   _RAM_BANK = 1;
   cbm_k_load(LOAD_TO_X16_RAM, 0xA000);
   _RAM_BANK = 0;
}

// ============= VERA info and my some of my own experimental results =============

// I've been experimenting with VERA a bit now and here are my conclusion and solutions to problems I incountered so far:

// Conclusions: there seem to be several concerns when dealing with VERA and its graphics/bitmap mode:
//    - a limitation on the number of times you can 'call' VERA.data0 (like some internal byte-number overflows: 255 seems to be max. This 'counter' seems to be reset when addres_hi is set. Weird.)
//    - the fact that you overrun the 65535-VRAM address -> you need to set address_hi to 1 then
//    - it is important to set the Tile Height and Tile Width bits to 0 in bitmap mode. Otherwise you will experience some kind of 'tiling'.
//    - layer 1 has a *mapbase* address of 0, while layer 0 has a default *tilebase* address of 0. They will interfere (weirdly) if not set otherwise.
//    - don't forget to turn off layer 1 -> its on top of layer 0 ;)
//    - layer 1 has a *tilebase* address of $F800, while layer 0 (when 320x240) needs quire a bit more than 64k (>$FFFF)of memory. this won't fit, without disabling or *moving* layer 1
//    - So its probably best not to use address $10000 as tilebase, otherwise you run into the sprites-data and run out of memory (with 320x240 and 8bpp)
//    - mapbase doesnt affect anything in bitmap mode, no need to set it

// Solutions:
//    - we want to preserve layer 1 (to be able to turn on logging temporarily)
//    - we move the tilebase of layer 1 from $F800 to $5800
//    - we use this tilebase for layer 0: $6000 = 24576 = 65536 - 40960. And 40960 = 128 * 320 (about half the screen). This means: if y > 128 then address_hi should be 1. Thats easy to check.
//    - we don't do VERA.data0 more than ~250 times. So we use vertical lines (or limit to certain maxmimum horizontal lines) so we don't have to call it that much in a row.

// Other (sometimes handy) info:

// kernal FB functions: C:\x16emu_win-r38\x16-rom\kernal\drivers\x16\framebuffer.s
// kernal GRAPH functions: C:\x16emu_win-r38\x16-rom\kernal\graph\graph.s

// extern function in asm, available in c: https://github.com/SlithyMatt/x16-hello-cc65/blob/master/mixed/hello.asm
// calling the kernal from asm: https://github.com/SlithyMatt/x16-wireframe/blob/master/model.asm

// calling the kernal from p8: https://github.com/irmen/prog8/blob/master/examples/cx16/cobramk3-gfx.p8

// Also read: https://www.8bitcoding.com/p/tiles.html


void move_text_screen_in_memory() {
    u16 n;
    u8 data;
    u8 tile_width_and_height;
    u16 tile_base, new_tile_base;
    
    // We need to keep the old text screen intact. So we need to copy the tilebase that is at $F800 to $5800.
    tile_base = (VERA.layer1.tilebase & ~3) << 9; // = 0x0F800
    tile_width_and_height = VERA.layer1.tilebase & 3;
    new_tile_base = 0x5800;
    
    VERA.address_hi = 0x00 | (1 << 4); // setting bit 16 of vram address to the highest bit in the tilebase (=0), setting auto-increment value to 1
    for (n = 0; n < 0x0800; n++) {
        // FIXME: this is SLOW. We should use the two build-in auto-increments instead!
        VERA.address = tile_base + n;
        data = VERA.data0;
        VERA.address = new_tile_base + n;
        VERA.data0 = data;
    }
    
    VERA.layer1.tilebase = (new_tile_base >> 9) | tile_width_and_height; // set new tilebase for layer 1 (0x05800)
}

void make_layer_1_transparent () {
    u8 x, y;
    
    for (y = 0; y < 31; y++) {  // FIXME: we should not do more than 30 rows, but since we are not really in 40x30 (and do a vscroll) we have to do this now.
        for (x = 0; x < 40; x++) {
            set_character_tile_at_xy(32, x, y);
        }
    }
}

void draw_text(char * text, u16 x, u16 y, u8 color) {
    u8 ch, nr = 0;
    while (text[nr]) {
        if (text[nr] >= 'a' && text[nr] <= 'z') {
            ch = text[nr]-64;
        }
        /*
        else if (text[nr] >= '0' && text[nr] <= '9') {
            ch = text[nr]-32;
        }
        */
        else {
            ch = text[nr];
        }
        set_character_at_xy(ch, x + (nr*8), y, color);
        nr++;
    }
}

// This uses actual tiles in layer 1
void set_character_tile_at_xy(u8 character, u16 x, u16 y) {
    VERA.address_hi = 0x00 | (1 << 4); // setting bit 16 of vram address to 0, setting auto-increment value to 1
    VERA.address = 0x0000 + (y * 128 + x)*2; // FIXME: apparently we are (still) in 128 character mode
    VERA.data0 = character;
    VERA.data0 = 0x01; // FIXME: background color (0) vs foreground color (1)
}

// This copies character pixels and puts them in layer 0
// https://www.commanderx16.com/forum/index.php?/topic/275-user-defined-characters/
void set_character_at_xy(u8 character, u16 x, u16 y, u8 color) {
    u8 pixel_x, pixel_y;
    u8 hor_line_packed;
    u8 dummy;
    
    for (pixel_y = 0; pixel_y < 8; pixel_y++) {
        // Note that we move the tilebase of layer 1 to $5800
        VERA.control = 0x01; // ADDRSEL=1
        VERA.address = 0x5800 + character * 8 + pixel_y; 
        
        VERA.control = 0x00; // ADDRSEL=0
        if (y+pixel_y >= 128) {
            VERA.address_hi = 0x01 | (1 << 4); // setting bit 16 of vram address to 1, setting auto-increment value to 1
            VERA.address = 0x0000 + (y-128+pixel_y)*320 + x; 
        }
        else {
            VERA.address_hi = 0x00 | (1 << 4); // setting bit 16 of vram address to 0, setting auto-increment value to 1
            VERA.address = 0x6000 + (y+pixel_y)*320+ x; 
        }
        hor_line_packed = VERA.data1;
        
        for (pixel_x = 0; pixel_x < 8; pixel_x++) {
            if (hor_line_packed & 0x80) {
                VERA.data0 = color;
            }
            else {
                // We have to increment the addess for data0, but we don't want to write to the pixel, so we read from it
                dummy = VERA.data0;
            }
            hor_line_packed = hor_line_packed << 1;
        }
    }
}

void set_palette_color(u8 index, u8 red, u8 green, u8 blue) {
    // Palette starts at: $1FA00
    VERA.address_hi = 0x01 | (1 << 4); // setting bit 16 of vram address to 1, setting auto-increment value to 1
    VERA.address = 0xFA00 + index * 2;
    VERA.data0 = green << 4 | blue & 0x0F;
    VERA.data0 = red & 0x0F;
}

void enable_display_layer(u8 layer) {
    if (layer) {
        VERA.display.video = VERA.display.video | 32; // enable layer 1  (the default for x16)
    }
    else {
        VERA.display.video = VERA.display.video | 16; // enable layer 0
    }
}

void disable_display_layer(u8 layer) {
    if (layer) {
        VERA.display.video = VERA.display.video & ~32; // disable layer 1  (the default for x16)
    }
    else {
        VERA.display.video = VERA.display.video & ~16; // disable layer 0
    }
}

u8 display_layer_is_enabled(u8 layer) {
    if (layer) {
        return VERA.display.video & 32;
    }
    else {
        return VERA.display.video & 16;
    }
}

void init_draw_screen() {
    VERA.display.vscale = 0x40; // scale / 2 (320px)
    VERA.display.hscale = 0x40; // scale / 2 (240px)

    VERA.layer0.config = (4 + 3); // enable bitmap mode and color depth = 8pp on layer 0
    VERA.layer0.tilebase = (0x0060 >> 1) | 0; // set new tilebase for layer 0 (0x06000)
    VERA.layer0.hscroll = 0; // This will clear the pallete offset
}

/*
void generate_render_code() {
    u16 n, j;
    u16 capped_height;
    u16 height;
    u8 y_pixel, previous_y_pixel;
    u8 nr_of_loads_to_add;
    
    // TODO: we want to produce draw-code for all possible heights of a wall. This height can exceed the RENDER_HEIGHT, if you stand very close to it.
    //       we now generate code for all heights up to 511. This might not be enough. We need to either cap this height (when drawing)
    //       or make sure this height is never asked for: simply by not being able to stand so close to a wall. That maximum height possible should be used here.
    
    // NOTE: the emulator only has 64 banks of 8kb (=512kb) !!
    // NOTE: be careful, printf will change the RAM_BANK!!
    // printf("height, n, y_pixel: %d, %d, %d\n", height, n, y_pixel);

    // FIXME: we start at height = 8 because 0-7 cannot be used because bank 0 is occupied/used by the kernal. But we should be sure no wall of height 0-7 is ever drawn.
    for (height = 8; height < 512; height++) {

        if (height > RENDER_HEIGHT) {
            capped_height = RENDER_HEIGHT;
        }
        else {
            capped_height = height;
        }
        
        _RAM_BANK = height/8;
        byte_code_index = (height%8) * 1024; // sub bank = 1024 bytes
        
        // We assume we always load at least once.
        // add_load_instruction();
        _banked_ram[byte_code_index++] = 0xAD; // lda ....
        _banked_ram[byte_code_index++] = 0x24; // VERA_DATA1
        _banked_ram[byte_code_index++] = 0x9F; 
        // We compare with previous_y_pixel to check whether we should add loads. We start at 0 (not -1) because we always load at least once (see above)
        previous_y_pixel = 0;
        for (n = 0; n < capped_height ; n++) {
            // y_pixel = pick_texture_pixel_row(n, capped_height, height);
            // WARNING: this is copied from pick_texture_pixel_row()!
            // we use the *original* (non-capped-to-RENDER_HEIGHT) height to determine which pixel to pick
            if (capped_height < height) {
                y_pixel = ((n+(height-capped_height)/2)*64)/height;
            }
            else {
                y_pixel = (n*64)/height;
            }
            
            // We use previous_y_pixel to determine whether we should produce LDA or STA opcodes (and how many)
            if (y_pixel == previous_y_pixel) {
                // No need to load, since we have to store using the same data as before
                // add_store_instruction();
                _banked_ram[byte_code_index++] = 0x8D; // sta ....
                _banked_ram[byte_code_index++] = 0x23; // VERA_DATA0
                _banked_ram[byte_code_index++] = 0x9F; 
            }
            else if (y_pixel > previous_y_pixel) {
                nr_of_loads_to_add = y_pixel - previous_y_pixel;
                for (j = 0; j < nr_of_loads_to_add; j++) {
                    // add_load_instruction();
                    _banked_ram[byte_code_index++] = 0xAD; // lda ....
                    _banked_ram[byte_code_index++] = 0x24; // VERA_DATA1
                    _banked_ram[byte_code_index++] = 0x9F; 
                }
                // add_store_instruction();
                _banked_ram[byte_code_index++] = 0x8D; // sta ....
                _banked_ram[byte_code_index++] = 0x23; // VERA_DATA0
                _banked_ram[byte_code_index++] = 0x9F; 
            }
            else {
                printf("ERROR: y_pixel is smaller than previous_y_pixel!\n");
                while(1);
            }
            previous_y_pixel = y_pixel;
        }
        // add_rts_instruction();
        _banked_ram[byte_code_index++] = 0x60; // rts
        
        // TODO: we do height - 2 , because we start at 4
        draw_progress_piece(height-2);
        
    }
    
    // setting back to bank 0
    _RAM_BANK = 0;
}
*/

// FIXME: this doesn't pick the last pixels (62 and 63), right? we should investigate this...
// WARNING: the content of this function is duplicated in generate_render_code!
u8 pick_texture_pixel_row(u16 point_on_capped_height, u16 capped_height, u16 height) {
    u8 y_pixel;
    // we use the *original* (non-capped-to-RENDER_HEIGHT) height to determine which pixel to pick
    if (capped_height < height) {
        y_pixel = ((height-(point_on_capped_height+(height-capped_height)/2))*64)/height;
    }
    else {
        y_pixel = ((height-point_on_capped_height)*64)/height;
    }
    return y_pixel;
}

void copy_texture_to_vram (u8 texture_index) {
    u16 x_pixel, y_pixel;
    u8 number_of_palette_indexes, color_index;
    
    _RAM_BANK = 1;
    
    number_of_palette_indexes = _banked_ram[4096]/3; // each palette index takes 3 bytes, so we divide by 3
    for (color_index = 0; color_index < number_of_palette_indexes; color_index++) {
        set_palette_color(color_index + palette_index, 
                          _banked_ram[4096+1+ color_index*3 + 2]>>4, 
                          _banked_ram[4096+1+ color_index*3 + 1]>>4, 
                          _banked_ram[4096+1+ color_index*3 + 0]>>4);
    }
    
    VERA.address_hi = 0x01 | (1 << 4); // setting bit 16 of vram address to the highest bit of the texture address (=1), setting auto-increment value to 1 (note: 1=1)
    VERA.address = TEXTURES_VRAM_BASE + texture_index * 4096; // TEXTURES_VRAM_BASE is 18C00 right now, right behind the BITMAP memory
    for (y_pixel = 0; y_pixel < 64; y_pixel++) {
        for (x_pixel = 0; x_pixel < 64; x_pixel++) {
            VERA.data0 = _banked_ram[y_pixel*64+x_pixel] + palette_index;
        }
    }
    _RAM_BANK = 0;
    
    palette_index += number_of_palette_indexes; // We increase the palette_index so the next texture will begin where this one ended
}

void draw_vertical_bar(u16 x_start, u16 height, u8 color) {
    u16 n;
    u16 start_y, end_y;
    u8 pixel_color;
    u16 tile_base = ((VERA.layer0.tilebase & ~3) << 9);
    u16 x_pixel, y_pixel;
    u16 capped_height;
    
    if (height > RENDER_HEIGHT) {
        capped_height = RENDER_HEIGHT;
    }
    else {
        capped_height = height;
    }
    
    start_y = TOP_BORDER_HEIGHT + RENDER_HEIGHT / 2 - capped_height / 2;
    end_y = start_y + capped_height;
    
    VERA.address_hi = 0x00 | (14 << 4); // setting bit 16 of vram address to the highest bit in the tilebase (=0), setting auto-increment value to 320 (note: 14=320)
    VERA.address = (tile_base + TOP_BORDER_HEIGHT*SCREEN_WIDTH + x_start) & 0xffff;
    for (n = start_y - TOP_BORDER_HEIGHT; n > 0; ) {
        if (n > 16) {
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            n = n - 16;
        }
        else if (n > 8) {
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            VERA.data0 = CEILING_COLOR;
            n = n - 8;
        }
        else {
            VERA.data0 = CEILING_COLOR;
            n--;
        }
    }
    pixel_color = color;
    // FIXME: we now switch between bank 1 and 0 each vertical line.
    _RAM_BANK = 1;
    for (n = capped_height; n > 0;) {
        if(use_textures) {
            
            // FIXME: we interpret 'color' here as the x-offset in the texture (color>>2 ranges from 0 to 63)
            x_pixel = color/4;
            y_pixel = pick_texture_pixel_row(n, capped_height, height);

            // FIXME: we only pick from one texture, since we are not using VRAM here for textures.
            pixel_color = _banked_ram[y_pixel*64+x_pixel] + 8*16;
            VERA.data0 = pixel_color;
            n--;
        }
        else {
            if (n > 16) {
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                n = n - 16;
            }
            else if (n > 8) {
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                VERA.data0 = pixel_color;
                n = n - 8;
            }
            else {
                VERA.data0 = pixel_color;
                n--;
            }
        }
    }
    _RAM_BANK = 0;
    for (n = TOP_BORDER_HEIGHT + RENDER_HEIGHT - end_y; n > 0; ) {
        if (n > 16) {
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            n = n - 16;
        }
        else if (n > 8) {
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            VERA.data0 = FLOOR_COLOR;
            n = n - 8;
        }
        else {
            VERA.data0 = FLOOR_COLOR;
            n--;
        }
    }
    
}

/*
void draw_progress_bar_border () {
    u16 x, y, n;
    u16 tile_base = ((VERA.layer0.tilebase & ~3) << 9);
    u8 color;
    i16 progress_index;
    
    y = (SCREEN_HEIGHT / 2) - PROGRESS_BAR_BORDER_HEIGHT - PROGRESS_BAR_HEIGHT/2;
    color = 6;

    for (progress_index = -5; progress_index < 256 + 5; progress_index++) {
        x = progress_index + (RENDER_WIDTH - 256)/2;
        
        VERA.address_hi = 0x00 | (14 << 4); // setting bit 16 of vram address to the highest bit in the tilebase (which is 0), setting auto-increment value to 320 (note: 14=320)
        VERA.address = (tile_base + y*SCREEN_WIDTH + x) & 0xffff;
        
        for (n = 0; n < PROGRESS_BAR_BORDER_HEIGHT*2 + PROGRESS_BAR_HEIGHT; n++) {
            if (progress_index <= 0 || progress_index > 256) {
                VERA.data0 = color;
            }
            else {
                if (n < PROGRESS_BAR_BORDER_HEIGHT || n >= PROGRESS_BAR_BORDER_HEIGHT + PROGRESS_BAR_HEIGHT) {
                    VERA.data0 = color;
                }
                else {
                    VERA.data0 = 16;
                }
            }
        }
    }
    
}
*/

/*
void draw_progress_piece (u16 progress_index) {
    u16 x, y, n;
    u16 tile_base = ((VERA.layer0.tilebase & ~3) << 9);
    u8 color;
    
    x = progress_index/2 + (RENDER_WIDTH - 256)/2;
    y = (SCREEN_HEIGHT / 2) - PROGRESS_BAR_HEIGHT/2;
    color = 2;

    VERA.address_hi = 0x00 | (14 << 4); // setting bit 16 of vram address to the highest bit in the tilebase (which is 0), setting auto-increment value to 320 (note: 14=320)
    VERA.address = (tile_base + y*SCREEN_WIDTH + x) & 0xffff;
    
    for (n = 0; n < PROGRESS_BAR_HEIGHT; n++) {
        VERA.data0 = color;
    }
    
}
*/

void clear_screen () {
    u16 x, y;
    u16 tile_base = ((VERA.layer0.tilebase & ~3) << 9);
    
    for (x = 0; x < SCREEN_WIDTH; x++) {
        
        // TODO: we should actually get the highest bit from VERA.layer0.tilebase 
        VERA.address_hi = 0x00 | (14 << 4); // setting bit 16 of vram address to the highest bit in the tilebase (which is 0), setting auto-increment value to 320 (note: 14=320)
        VERA.address = (tile_base + x) & 0xffff;
        // Full screen is black
        for (y = 0; y < ((PAL_SCREEN_HEIGHT)/8); y++) {
            VERA.data0 = 16;
            VERA.data0 = 16;
            VERA.data0 = 16;
            VERA.data0 = 16;
            VERA.data0 = 16;
            VERA.data0 = 16;
            VERA.data0 = 16;
            VERA.data0 = 16;
        }
    }
}

void clear_render_area () {
    u16 x, y;
    u16 tile_base = ((VERA.layer0.tilebase & ~3) << 9);
    u8 color;
    
    for (x = BORDER_WIDTH; x < BORDER_WIDTH + RENDER_WIDTH; x++) {
        
        VERA.address_hi = 0x00 | (14 << 4); // setting bit 16 of vram address to the highest bit in the tilebase (which is 0), setting auto-increment value to 320 (note: 14=320)
        VERA.address = (tile_base + x + TOP_BORDER_HEIGHT*320) & 0xffff;
        
        color = 16;
        y = RENDER_HEIGHT;
        for (; y >= 8; y -= 8) {
            VERA.data0 = 16;
            VERA.data0 = 16;
            VERA.data0 = 16;
            VERA.data0 = 16;
            VERA.data0 = 16;
            VERA.data0 = 16;
            VERA.data0 = 16;
            VERA.data0 = 16;
        }
        for (; y > 0; y--) {
            VERA.data0 = 16;
        }
    }
}

void draw_background (u8 background_color, u8 stats_background_color) {
    u16 x, y;
    u16 tile_base = ((VERA.layer0.tilebase & ~3) << 9);
    u8 color;
    u8 draw_border = 1;
    
    /*
    VERA.control = 0x02; // DSEL=1
    VERA.display.vstart = 20;
    VERA.display.vstop = 220-1;
    VERA.control = 0x00; // DSEL=0
    */

    // We want to fill the entire buffer with one color. We use the auto-increment to do this faster (and set it to 320).
    
    for (x = 0; x < SCREEN_WIDTH; x++) {
        
        // TODO: we should actually get the highest bit from VERA.layer0.tilebase 
        VERA.address_hi = 0x00 | (14 << 4); // setting bit 16 of vram address to the highest bit in the tilebase (which is 0), setting auto-increment value to 320 (note: 14=320)
        VERA.address = (tile_base + x) & 0xffff;

        for (y = 0; y < TOP_BORDER_HEIGHT-1; y++) {
            VERA.data0 = background_color;
        }

        // Top part of screen is given primary background_color
        if (x < BORDER_WIDTH-1) {
            color = background_color;
        }
        else if (x == BORDER_WIDTH-1) {
            color = 16; // left dark line
        }
        else if (x >= BORDER_WIDTH && x < BORDER_WIDTH + RENDER_WIDTH) {
            color = 16;
        }
        else if (x == BORDER_WIDTH + RENDER_WIDTH) {
            color = 167; // right dark green line
        }
        else {
            color = background_color;
        }
        
        VERA.data0 = color; // top dark line
        
        y = RENDER_HEIGHT;
        for (; y >= 8; y -= 8) {
            VERA.data0 = color;
            VERA.data0 = color;
            VERA.data0 = color;
            VERA.data0 = color;
            VERA.data0 = color;
            VERA.data0 = color;
            VERA.data0 = color;
            VERA.data0 = color;
        }
        for (; y > 0; y--) {
            VERA.data0 = color;
        }
        
        if (color == 16) {
            VERA.data0 = 167;  // bottom light green line
        }
        else {
            VERA.data0 = color; 
        }
        
        for (y = 0; y < MIDDLE_BORDER_HEIGHT-2; y++) {
            VERA.data0 = background_color;
        }
        
        // Part of the stats
        if (x < BORDER_WIDTH) {
            color = background_color;
        }
        else if (x == BORDER_WIDTH) {
            color = 16; // left dark line
        }
        else if (x == BORDER_WIDTH+1) {
            color = 192; // left light blue line
        }
        else if (x > BORDER_WIDTH+1 && x < BORDER_WIDTH + RENDER_WIDTH - 1) {
            color = stats_background_color;
        }
        else if (x == BORDER_WIDTH + RENDER_WIDTH - 1) {
            color = 195; // right dark blue line
        }
        else if (x == BORDER_WIDTH + RENDER_WIDTH) {
            color = 167; // right light green line
        }
        else {
            color = background_color;
        }
        
        if (color == stats_background_color || color == 192 || color == 195) {
            VERA.data0 = 0;  // top dark line
        }
        else {
            VERA.data0 = color;
        }
        if (color == stats_background_color) {
            VERA.data0 = 192;  // top light blue line
        }
        else {
            VERA.data0 = color;
        }
        // stats_background_color is blue
        // TODO: made it fit by adding -3 here. It should be properly add up.
        for (y = 0; y < ((SCREEN_HEIGHT-RENDER_HEIGHT-TOP_BORDER_HEIGHT-MIDDLE_BORDER_HEIGHT-BOTTOM_BORDER_HEIGHT-3)); y++) {
            VERA.data0 = color;
        }
        
        if (color == stats_background_color || color == 192 || color == 195) {
            VERA.data0 = 195;  // bottom dark blue line
        }
        else {
            VERA.data0 = color;
        }
        if (color == stats_background_color || color == 192 || color == 195 || color == 16) {
            VERA.data0 = 167; // bottom light green line
        }
        else {
            VERA.data0 = color;
        }
        
        for (y = 0; y < BOTTOM_BORDER_HEIGHT; y++) {
            VERA.data0 = background_color;
        }
    }
    
}

/*
char log_buffer[100];

void print_vera_info() {
    
    printf("general\n");
    printf("\n");
    printf("addr_hi:  %08s\n", itoa(VERA.address_hi, log_buffer, 2));
    printf("control:  %08s\n", itoa(VERA.control, log_buffer, 2));
    printf("video:    %08s\n", itoa(VERA.display.video, log_buffer, 2));
    printf("hscale:   %08s\n", itoa(VERA.display.hscale, log_buffer, 2));
    printf("vscale:   %08s\n", itoa(VERA.display.vscale, log_buffer, 2));
    printf("\n");
    
    printf("layer 0\n");
    printf("\n");
    printf("config:   %08s\n", itoa(VERA.layer0.config, log_buffer, 2));
    printf("tilebase: $%03x00 (%02s)\n", (VERA.layer0.tilebase & ~3) << 1 , itoa(VERA.layer0.tilebase & 3, log_buffer, 2));
    printf("mapbase:  $%03x00\n", VERA.layer0.mapbase << 1);
    printf("\n");
    
    printf("layer 1\n");
    printf("\n");
    printf("config:   %08s\n", itoa(VERA.layer1.config, log_buffer, 2));
    printf("tilebase: $%03x00 (%02s)\n", (VERA.layer1.tilebase & ~3) << 1 , itoa(VERA.layer1.tilebase & 3, log_buffer, 2));
    printf("mapbase:  $%03x00\n", VERA.layer1.mapbase << 1);
    printf("\n");
}
*/


/* From cx16.h: A structure with the Video Enhanced Retro Adapter's external registers */
/*  
struct __vera {
    unsigned short      address;        // Address for data ports 
    unsigned char       address_hi;
    unsigned char       data0;          // Data port 0 
    unsigned char       data1;          // Data port 1 
    unsigned char       control;        // Control register 
    unsigned char       irq_enable;     // Interrupt enable bits 
    unsigned char       irq_flags;      // Interrupt flags 
    unsigned char       irq_raster;     // Line where IRQ will occur 
    union {
        struct {                        // Visible when DCSEL flag = 0 
            unsigned char video;        // Flags to enable video layers 
            unsigned char hscale;       // Horizontal scale factor 
            unsigned char vscale;       // Vertical scale factor 
            unsigned char border;       // Border color (NTSC mode) 
        };
        struct {                        // Visible when DCSEL flag = 1 
            unsigned char hstart;       // Horizontal start position 
            unsigned char hstop;        // Horizontal stop position 
            unsigned char vstart;       // Vertical start position 
            unsigned char vstop;        // Vertical stop position 
        };
    } display;
    struct {
        unsigned char   config;         // Layer map geometry 
        unsigned char   mapbase;        // Map data address 
        unsigned char   tilebase;       // Tile address and geometry 
        unsigned int    hscroll;        // Smooth scroll horizontal offset 
        unsigned int    vscroll;        // Smooth scroll vertical offset 
    } layer0;
    struct {
        unsigned char   config;
        unsigned char   mapbase;
        unsigned char   tilebase;
        unsigned int    hscroll;
        unsigned int    vscroll;
    } layer1;
    struct {
        unsigned char   control;        // PCM format 
        unsigned char   rate;           // Sample rate 
        unsigned char   data;           // PCM output queue 
    } audio;                            // Pulse-Code Modulation registers 
    struct {
        unsigned char   data;
        unsigned char   control;
    } spi;                              // SD card interface 
};
*/
// #define VERA    (*(volatile struct __vera *)0x9F20)